const path = require('path');
const STYLE_EXTENSIONS = ['.css', '.scss', '.sass', '.pcss', '.stylus', '.styl', '.less', '.sss'];
const ASSET_EXTENSIONS = ['.gif', '.jpeg', '.jpg', '.png', '.svg', '.mp4', '.webm', '.ogv'];

/**
 * like: "var a = require('./bg.img')"
 * @param {import('@babel/core').NodePath<import('@babel/types').CallExpression>} nodePath
 */
function checkBeforeRemoveRequire(nodePath) {
  if (nodePath.parentPath.isVariableDeclarator()) {
    throw new Error(`${requirePath} should not be assigned to variable.`);
  }
}

/**
 * like: "import a from './test.less'"
 * @param {import('@babel/core').NodePath<import('@babel/types').ImportDeclaration>} nodePath
 */
function checkBeforeRemoveImport(nodePath) {
  const specifiers = nodePath.get('specifiers');

  if (specifiers.length) {
    const specifier = specifiers[specifiers.length - 1];

    if (specifier.isImportDefaultSpecifier()) {
      throw new Error(`${nodePath.node.source.value} should not be imported using default imports.`);
    }
    if (specifier.isImportSpecifier()) {
      throw new Error(`${nodePath.node.source.value} should not be imported using named imports.`);
    }
    if (specifier.isImportNamespaceSpecifier()) {
      throw new Error(`${nodePath.node.source.value} should not be imported using namespace imports.`);
    }
  }
}

/**
 *
 * @param {string} filePath
 * @param {((path: string) => string) | boolean | string | undefined} replacer
 */
function handleReplaceStyleImport(filePath, replacer) {
  const ext = path.extname(filePath);
  const target =
    typeof replacer === 'function'
      ? replacer(filePath)
      : replacer === true
      ? filePath.replace(ext, '.css')
      : typeof replacer === 'string'
      ? filePath.replace(ext, replacer)
      : undefined;
  return target;
}

/**
 * TODO: inspect replace file path exist
 * @param {string} targetPath
 * @param {*} nodePath
 */
function inspectExist(targetPath, nodePath) {
  return true;
  // try {
  //   const { filename: baseFilePath, cwd } = nodePath.hub.file.opts;
  //   const absPath = path.join(path.dirname(baseFilePath), targetPath);
  //   return fs.existsSync(absPath);
  // } catch (error) {
  //   console.debug('error:', error);
  //   return true;
  // }
}

/** @type {(api: { types: typeof import('@babel/core').types}) => import('@babel/core').PluginObj} */
const transformer = ({ types }) => {
  /** @type {import('@babel/core').PluginObj} */
  const pluginObj = {
    visitor: {
      CallExpression: {
        enter(nodePath, { opts }) {
          const callee = nodePath.get('callee');
          const { removeAssetImport = true, removeStyleImport = false, replaceStyleImport = '.css' } = opts || {};

          // like: "require('xxx')"
          if (callee.isIdentifier() && callee.equals('name', 'require')) {
            const arg = nodePath.get('arguments')[0];

            if (arg && types.isStringLiteral(arg)) {
              const requirePath = arg.node.value;

              if (requirePath && typeof requirePath === 'string') {
                const ext = path.extname(requirePath);
                const fileName = path.basename(requirePath);
                if (arg && arg.isStringLiteral()) {
                  // handle require asset
                  if (ASSET_EXTENSIONS.includes(ext)) {
                    if (removeAssetImport) {
                      checkBeforeRemoveRequire(nodePath);
                      nodePath.remove();
                    }
                  }

                  // handle require style
                  if (STYLE_EXTENSIONS.includes(ext)) {
                    if (removeStyleImport) {
                      checkBeforeRemoveRequire(nodePath);
                      nodePath.remove();
                    } else if (replaceStyleImport) {
                      const replaceTarget = handleReplaceStyleImport(requirePath, replaceStyleImport);
                      if (typeof replaceTarget === 'string' && inspectExist(replaceTarget, nodePath)) {
                        arg.node.value = replaceTarget;
                      }
                    }
                  }
                }
              }
            }
          }
        },
      },

      ImportDeclaration: {
        enter(nodePath, { opts }) {
          const importPath = nodePath.node.source.value;

          if (importPath && typeof importPath === 'string') {
            const ext = path.extname(importPath);
            const fileName = path.basename(importPath);
            const { removeAssetImport = true, removeStyleImport = false, replaceStyleImport = '' } = opts || {};

            // handle import asset
            if (ASSET_EXTENSIONS.includes(ext)) {
              if (removeAssetImport) {
                checkBeforeRemoveImport(nodePath);
                nodePath.remove();
              }
            }

            // handle import style
            if (STYLE_EXTENSIONS.includes(ext)) {
              if (removeStyleImport) {
                checkBeforeRemoveImport(nodePath);
                nodePath.remove();
              } else if (replaceStyleImport) {
                const replaceTarget = handleReplaceStyleImport(importPath, replaceStyleImport);
                if (typeof replaceTarget === 'string' && inspectExist(replaceTarget, nodePath)) {
                  nodePath.node.source.value = replaceTarget;
                }
              }
            }
          }
        },
      },
    },
  };

  return pluginObj;
};

module.exports = transformer;
module.exports.default = transformer;
